(**************************************************************************

   	This file is part of the AtomicDensityMatrix package for Mathematica
    Copyright (C) 2004-2014 Simon M. Rochester <simon@rochesterscientific.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

**************************************************************************)


BeginPackage["AtomicDensityMatrix`DensityMatrix`",
    { "AtomicDensityMatrix`",
        "AtomicDensityMatrix`Common`",
        "AtomicDensityMatrix`AtomicSystem`",
        "AtomicDensityMatrix`ExperimentalSystem`"}]

Begin["`Private`"] (* Begin Private Context *)

(* ::Subsection:: *)
(*Density matrix*)

Options[DensityMatrix] = {
    TimeDependence -> True,
    Representation -> "Zeeman",
    DMSymbol -> \[Rho],
    DMLabel -> None,
    ComplexExpandVariables -> False,
    TimeVariable -> t};

Options[Ket] = {TimeDependence -> Automatic, TimeVariable -> Automatic, ComplexExpandVariables -> Automatic};

Ket[syst : {_AtomicState..}, symbol_, opts : OptionsPattern[]] := Module[{ket, newopts},
    newopts = Join[DeleteCases[Join[{opts}, Options[Ket]], _ -> Automatic], Options[DensityMatrix]];
    ket = Switch[ComplexExpandVariables /. newopts,
        True,
        Re[#]+I Im[#]& /@ (Subscript[symbol, SublevelLabel[#]]& /@ syst),
        Subscript,
        Subscript[symbol, Re, SublevelLabel[#]]+I Subscript[symbol, Im, SublevelLabel[#]]& /@ syst,
        _,
        Subscript[symbol, SublevelLabel[#]]& /@ syst
    ];
    If[TimeDependence /. newopts, ket = ket /. x : Subscript[symbol, __] -> x[t]];
    Transpose[{ket}]
]

Quiet[LevelLabelPattern[State_, StateLabel_, F_, M_] :=
    State : {StateLabel_, F_, M_}|{StateLabel_, M_}|StateLabel_?(Head[#] =!= List&), {RuleDelayed::"rhs"}]

LevelLabelPattern1[] := LevelLabelPattern[State1, StateLabel1, F1, M1]

LevelLabelPattern2[] := LevelLabelPattern[State2, StateLabel2, F2, M2]

(*LabelPattern[]=DMLabel_:None*)

Options[DMElementPattern] = {DMSymbol -> Automatic, TimeDependence -> Automatic, ComplexExpandVariables -> Automatic}

DMElementPattern[opts : OptionsPattern[]] := Module[{opts1, r},
    opts1 = Join[DeleteCases[Join[{opts}, Options[DMElementPattern]], _ -> Automatic], Options[DensityMatrix]];
    r = DMSymbol /. opts1;
    Switch[{ComplexExpandVariables, TimeDependence} /. opts1,
        {Subscript, False}, DMElementPartPattern[r],
        {Subscript, True}, DMElementPartTDPattern[r],
        {Subscript, All}, DMElementPartPattern[r]|DMElementPartTDPattern[r],
        {_, False}, DMElementPattern[r],
        {_, True}, DMElementTDPattern[r],
        {_, All}, DMElementPattern[r]|DMElementTDPattern[r]
    ]
]

DMElementPattern[r : Except[___Rule]] = (
    Subscript[r, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]]|
        Subscript[r, LevelLabelPattern1[], LevelLabelPattern2[]]
) /; (DMLabel =!= Re&&DMLabel =!= Im)

DMElementTDPattern[r : Except[___Rule]] = (
    Derivative[TimeDerivative_][Subscript[r, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]]][Time_]|
        Subscript[r, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]][Time_]|
        Derivative[TimeDerivative_][Subscript[r, LevelLabelPattern1[], LevelLabelPattern2[]]][Time_]|
        Subscript[r, LevelLabelPattern1[], LevelLabelPattern2[]][Time_]
) /; (DMLabel =!= Re&&DMLabel =!= Im)

DMElementPartPattern[r : Except[___Rule]] = (
    Subscript[r, ReImPart : Re|Im, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]]|
        Subscript[r, ReImPart : Re|Im, LevelLabelPattern1[], LevelLabelPattern2[]]
) /; (DMLabel =!= Re&&DMLabel =!= Im)

DMElementPartTDPattern[r : Except[___Rule]] = (
    Derivative[TimeDerivative_][Subscript[r, ReImPart : Re|Im, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]]][Time_]|
        Subscript[r, ReImPart : Re|Im, DMLabel_, LevelLabelPattern1[], LevelLabelPattern2[]][Time_]|
        Derivative[TimeDerivative_][Subscript[r, ReImPart : Re|Im, LevelLabelPattern1[], LevelLabelPattern2[]]][Time_]|
        Subscript[r, ReImPart : Re|Im, LevelLabelPattern1[], LevelLabelPattern2[]][Time_]
) /; (DMLabel =!= Re&&DMLabel =!= Im)

ClearAll[ExpandDMVariablesRule];

ExpandDMVariablesRule[syst_, opts___Rule] :=
    Module[
        {newopts, sll, patt, rep, imp, elem, rept, impt, rep1, imp1, rept1, impt1, elemt, e, r, i, r1, i1},
        newopts = Join[{opts}, Options[DensityMatrix]] /. (DMLabel -> (None|False)) -> (DMLabel -> Sequence[]);
        sll = SublevelLabel /@ syst;
        patt = DMElementPattern[ComplexExpandVariables -> False, TimeDependence -> All, opts];
        Switch[ComplexExpandVariables /. newopts,
            True,
            (v : patt :> Switch[
                e = If[{Time} == {}, elem, elemt];
                Order[Position[sll, State1, {1}], Position[sll, State2, {1}]],
                -1, Re[e]-I Im[e],
                0, Re[v],
                1, Re[v]+I Im[v]
            ]) /. {elem -> DMElement[State2, State1, opts],
                elemt -> DMElement[State2, State1, Time, TimeDerivative, opts]},
            Subscript,
            (v : patt :> Switch[
                {r, i, r1, i1} = If[{Time} == {}, {rep, imp, rep1, imp1}, {rept, impt, rept1, impt1}];
                Order[Position[sll, State1, {1}], Position[sll, State2, {1}]],
                -1, r1-I i1,
                0, r,
                1, r+I i
            ]) /. {rep -> DMElementPart[Re, State1, State2, opts],
                imp -> DMElementPart[Im, State1, State2, opts],
                rept -> DMElementPart[Re, State1, State2, Time, TimeDerivative, opts],
                impt -> DMElementPart[Im, State1, State2, Time, TimeDerivative, opts],
                rep1 -> DMElementPart[Re, State2, State1, opts],
                imp1 -> DMElementPart[Im, State2, State1, opts],
                rept1 -> DMElementPart[Re, State2, State1, Time, TimeDerivative, opts],
                impt1 -> DMElementPart[Im, State2, State1, Time, TimeDerivative, opts]},
            _,
            {}
        ]
    ]

ReImToSubscriptRule[opts___Rule] :=
    (f : Re|f : Im)[DMElementPattern[ComplexExpandVariables -> False, opts, TimeDependence -> All]] :>
        DMElementPart[f, State1, State2, Time, TimeDerivative, opts]

DMElement[s1_, s2_, opts___Rule] :=
    Module[{sym, lab},
        lab = DMLabel /. {opts} /. Options[DensityMatrix] /. (None|False -> Sequence[]);
        sym = DMSymbol /. {opts} /. Options[DensityMatrix];
        Subscript[sym, lab, s1, s2]
    ]

DMElement[s1_, s2_, time_, opts___Rule] := DMElement[s1, s2, opts][time]

DMElement[s1_, s2_, time_, d_, opts___Rule] := Derivative[d][DMElement[s1, s2, opts]][time]

DMElementPart[f_, s1_, s2_, opts___Rule] := Module[{sym, lab},
    lab = DMLabel /. {opts} /. Options[DensityMatrix] /. (None|False -> Sequence[]);
    sym = DMSymbol /. {opts} /. Options[DensityMatrix];
    Subscript[sym, f, lab, s1, s2]
]

DMElementPart[f_, s1_, s2_, time_, opts___Rule] := DMElementPart[f, s1, s2, opts][time]

DMElementPart[f_, s1_, s2_, time_, d_, opts___Rule] := Derivative[d][DMElementPart[f, s1, s2, opts]][time]

DensityMatrix[syst : {_AtomicState..}, opts : OptionsPattern[]] := Module[{cev, newopts, rfunc},
    newopts = Join[{opts}, Options[DensityMatrix]] /. (DMLabel -> (None|False)) -> (DMLabel -> Sequence[]);
    Switch[Representation /. newopts,
        "Zeeman",
        cev = ComplexExpandVariables /. newopts;
        rfunc[l1_, l2_] = If[TimeDependence /. newopts, DMElement[l1, l2, TimeVariable /. newopts, opts], DMElement[l1, l2, opts]];
        DirectSelfProduct[rfunc, SublevelLabel /@ syst] /. ExpandDMVariablesRule[syst, opts],
        PolarizationMoments,
        PolarizationMoments[syst, opts]
    ]
]

DensityMatrix[as : {_AtomicState..}, es : {_Region..}, opts : OptionsPattern[]] := DensityMatrix[as, DMLabel -> RegionLabel[#], opts]& /@ es

ExprToReIm[expr_List, opts___Rule] := Flatten[ExprToReIm[#, opts]& /@ expr]

ExprToReIm[expr_, opts___Rule] :=
    DeleteCases[
        Chop@Refine[
            ComplexExpand[
                Flatten@Through[{Re, Im}@expr] /. {(f : Re|Im)[x_ == y_] :> f[x] == f[y]},
                DMElementPattern[ComplexExpandVariables -> False, opts, TimeDependence -> All]
            ],
            DMElementPattern[ComplexExpandVariables -> Subscript, opts, TimeDependence -> All]\[Element]Reals
        ]
            /. {Re[x : DMElementPattern[ComplexExpandVariables -> Subscript, opts, TimeDependence -> All]] -> x,
            Im[x : DMElementPattern[ComplexExpandVariables -> Subscript, opts, TimeDependence -> All]] -> 0},
        True|0
    ]

SetAttributes[ERI, Listable]

ERI[x_ == y_] := MapThread[Equal, ERI@{x, y}]

ERI[x_] :=
    Total@Flatten@# & /@ Reap[ERISow[Expand[x /. y : Exp[_] :> ExpToTrig[y]]], {Re, Im}][[2]]

ERISow[x_Plus] := Scan[ERISow, x]

ERISow[x_] :=
    If[
        FreeQ[x, Complex],
        Sow[x, Re],
        If[
            MatchQ[x, Complex[0, a_] b_. /; FreeQ[b, Complex]],
            Sow[x/I, Im],
            Sow[ComplexExpand[Re[x], TargetFunctions -> {Re, Im}], Re];
            Sow[ComplexExpand[Im[x], TargetFunctions -> {Re, Im}], Im];
        ]
    ]

Options[DMVariables] = {
    TimeDependence -> Automatic,
    Representation -> Automatic,
    DMSymbol -> Automatic,
    DMLabel -> Automatic,
    ComplexExpandVariables -> Automatic,
    TimeVariable -> Automatic};

DMVariables[syst : {_AtomicState..}, opts___Rule] :=
    Module[{cev, newopts},
        newopts = DeleteCases[Join[{opts}, Options[DMVariables]], _ -> Automatic];
        {cev} = {ComplexExpandVariables} /. newopts /. Options[DensityMatrix];
        If[cev||cev === Subscript,
            DeleteCases[Flatten@ERI@UpperTriangle[#], 0]&,
            Flatten
        ]@DensityMatrix[syst, Sequence @@ newopts]
    ]

DMVariables[as : {_AtomicState..}, es : {_Region..}, opts___Rule] := Flatten[DMVariables[as, DMLabel -> RegionLabel[#], opts]& /@ es]

End[] (* End Private Context *)

EndPackage[]